@@ -18,14 +18,14 @@ annotations.
 import re
 from textwrap import dedent
 
-from parso import parse
+from parso import parse, ParserSyntaxError
 
 from jedi._compatibility import u
-from jedi.common import unite
-from jedi.evaluate import context
+from jedi.evaluate.utils import indent_block
 from jedi.evaluate.cache import evaluator_method_cache
-from jedi.common import indent_block
-from jedi.evaluate.iterable import SequenceLiteralContext, FakeSequence
+from jedi.evaluate.base_context import iterator_to_context_set, ContextSet, \
+    NO_CONTEXTS
+from jedi.evaluate.lazy_context import LazyKnownContexts
 
 
 DOCSTRING_PARAM_PATTERNS = [
@@ -42,49 +42,59 @@ DOCSTRING_RETURN_PATTERNS = [
 REST_ROLE_PATTERN = re.compile(r':[^`]+:`([^`]+)`')
 
 
-try:
-    from numpydoc.docscrape import NumpyDocString
-except ImportError:
-    def _search_param_in_numpydocstr(docstr, param_str):
-        return []
+_numpy_doc_string_cache = None
 
-    def _search_return_in_numpydocstr(docstr):
-        return []
-else:
-    def _search_param_in_numpydocstr(docstr, param_str):
-        """Search `docstr` (in numpydoc format) for type(-s) of `param_str`."""
-        try:
-            # This is a non-public API. If it ever changes we should be 
-            # prepared and return gracefully.
-            params = NumpyDocString(docstr)._parsed_data['Parameters']
-        except (KeyError, AttributeError):
-            return []
-        for p_name, p_type, p_descr in params:
-            if p_name == param_str:
-                m = re.match('([^,]+(,[^,]+)*?)(,[ ]*optional)?$', p_type)
-                if m:
-                    p_type = m.group(1)
-                return list(_expand_typestr(p_type))
-        return []
 
-    def _search_return_in_numpydocstr(docstr):
-        """
-        Search `docstr` (in numpydoc format) for type(-s) of function returns.
-        """
-        doc = NumpyDocString(docstr)
-        try:
-            # This is a non-public API. If it ever changes we should be 
-            # prepared and return gracefully.
-            returns = doc._parsed_data['Returns']
-            returns += doc._parsed_data['Yields']
-        except (KeyError, AttributeError):
-            raise StopIteration
-        for r_name, r_type, r_descr in returns:
-            #Return names are optional and if so the type is in the name
-            if not r_type:
-                r_type = r_name
-            for type_ in _expand_typestr(r_type):
-                yield type_
+def _get_numpy_doc_string_cls():
+    global _numpy_doc_string_cache
+    try:
+        from numpydoc.docscrape import NumpyDocString
+        _numpy_doc_string_cache = NumpyDocString
+    except ImportError as e:
+        _numpy_doc_string_cache = e
+    if isinstance(_numpy_doc_string_cache, ImportError):
+        raise _numpy_doc_string_cache
+    return _numpy_doc_string_cache
+
+
+def _search_param_in_numpydocstr(docstr, param_str):
+    """Search `docstr` (in numpydoc format) for type(-s) of `param_str`."""
+    try:
+        # This is a non-public API. If it ever changes we should be
+        # prepared and return gracefully.
+        params = _get_numpy_doc_string_cls()(docstr)._parsed_data['Parameters']
+    except (KeyError, AttributeError, ImportError):
+        return []
+    for p_name, p_type, p_descr in params:
+        if p_name == param_str:
+            m = re.match('([^,]+(,[^,]+)*?)(,[ ]*optional)?$', p_type)
+            if m:
+                p_type = m.group(1)
+            return list(_expand_typestr(p_type))
+    return []
+
+
+def _search_return_in_numpydocstr(docstr):
+    """
+    Search `docstr` (in numpydoc format) for type(-s) of function returns.
+    """
+    try:
+        doc = _get_numpy_doc_string_cls()(docstr)
+    except ImportError:
+        return
+    try:
+        # This is a non-public API. If it ever changes we should be
+        # prepared and return gracefully.
+        returns = doc._parsed_data['Returns']
+        returns += doc._parsed_data['Yields']
+    except (KeyError, AttributeError):
+        return
+    for r_name, r_type, r_descr in returns:
+        # Return names are optional and if so the type is in the name
+        if not r_type:
+            r_type = r_name
+        for type_ in _expand_typestr(r_type):
+            yield type_
 
 
 def _expand_typestr(type_str):
@@ -145,8 +155,7 @@ def _search_param_in_docstr(docstr, param_str):
         if match:
             return [_strip_rst_role(match.group(1))]
 
-    return (_search_param_in_numpydocstr(docstr, param_str) or
-            [])
+    return _search_param_in_numpydocstr(docstr, param_str)
 
 
 def _strip_rst_role(type_str):
@@ -179,7 +188,7 @@ def _evaluate_for_statement_string(module_context, string):
         Need this docstring so that if the below part is not valid Python this
         is still a function.
         '''
-    {0}
+    {}
     """))
     if string is None:
         return []
@@ -193,7 +202,10 @@ def _evaluate_for_statement_string(module_context, string):
     # will be impossible to use `...` (Ellipsis) as a token. Docstring types
     # don't need to conform with the current grammar.
     grammar = module_context.evaluator.latest_grammar
-    module = grammar.parse(code.format(indent_block(string)))
+    try:
+        module = grammar.parse(code.format(indent_block(string)), error_recovery=False)
+    except ParserSyntaxError:
+        return []
     try:
         funcdef = next(module.iter_funcdefs())
         # First pick suite, then simple_stmt and then the node,
@@ -202,7 +214,7 @@ def _evaluate_for_statement_string(module_context, string):
     except (AttributeError, IndexError):
         return []
 
-    from jedi.evaluate.representation import FunctionContext
+    from jedi.evaluate.context import FunctionContext
     function_context = FunctionContext(
         module_context.evaluator,
         module_context,
@@ -223,7 +235,10 @@ def _execute_types_in_stmt(module_context, stmt):
     contain is executed. (Used as type information).
     """
     definitions = module_context.eval_node(stmt)
-    return unite(_execute_array_values(module_context.evaluator, d) for d in definitions)
+    return ContextSet.from_sets(
+        _execute_array_values(module_context.evaluator, d)
+        for d in definitions
+    )
 
 
 def _execute_array_values(evaluator, array):
@@ -231,22 +246,26 @@ def _execute_array_values(evaluator, array):
     Tuples indicate that there's not just one return value, but the listed
     ones.  `(str, int)` means that it returns a tuple with both types.
     """
+    from jedi.evaluate.context.iterable import SequenceLiteralContext, FakeSequence
     if isinstance(array, SequenceLiteralContext):
         values = []
         for lazy_context in array.py__iter__():
-            objects = unite(_execute_array_values(evaluator, typ) for typ in lazy_context.infer())
-            values.append(context.LazyKnownContexts(objects))
-        return set([FakeSequence(evaluator, array.array_type, values)])
+            objects = ContextSet.from_sets(
+                _execute_array_values(evaluator, typ)
+                for typ in lazy_context.infer()
+            )
+            values.append(LazyKnownContexts(objects))
+        return {FakeSequence(evaluator, array.array_type, values)}
     else:
         return array.execute_evaluated()
 
 
 @evaluator_method_cache()
 def infer_param(execution_context, param):
-    from jedi.evaluate.instance import AnonymousInstanceFunctionExecution
+    from jedi.evaluate.context.instance import AnonymousInstanceFunctionExecution
 
     def eval_docstring(docstring):
-        return set(
+        return ContextSet.from_iterable(
             p
             for param_str in _search_param_in_docstr(docstring, param.name.value)
             for p in _evaluate_for_statement_string(module_context, param_str)
@@ -254,7 +273,7 @@ def infer_param(execution_context, param):
     module_context = execution_context.get_root_context()
     func = param.get_parent_function()
     if func.type == 'lambdef':
-        return set()
+        return NO_CONTEXTS
 
     types = eval_docstring(execution_context.py__doc__())
     if isinstance(execution_context, AnonymousInstanceFunctionExecution) and \
@@ -266,6 +285,7 @@ def infer_param(execution_context, param):
 
 
 @evaluator_method_cache()
+@iterator_to_context_set
 def infer_return_types(function_context):
     def search_return_in_docstr(code):
         for p in DOCSTRING_RETURN_PATTERNS:
@@ -279,4 +299,3 @@ def infer_return_types(function_context):
     for type_str in search_return_in_docstr(function_context.py__doc__()):
         for type_eval in _evaluate_for_statement_string(function_context.get_root_context(), type_str):
             yield type_eval
-
